﻿import graphical.CameraView;
import util.*;

/**
 * This Singleton Observer pattern based class provides the basic behavior for a 3Dscene. 
 * All it really does is accept objects that have x, y and x properties, and convert these 
 * to 2D values and drop these values on the object's _x and _y. 
 * A 3DScene has a camera object, which is used to translate 3D values to 2D values.
 *
 * This class is meant to be Abstract, and should not used directly. Instead,
 * inherit from this class and override the method createObjects(). All you need
 * to do is add the objects you want in the 3Dscene in this method.
 * An alternative method is to add and remove objects dynamically using add3DObject()
 * and remove3DObject().
 */
class graphical.AScene3D extends MovieClip implements Observer {
	
	// A static reference to the 3DScene instance (Singleton).
  	private static var scene3D:AScene3D = null;
	
	// Only one camera allowed for a 3DScene. I'll maybe add more in the future.
	private static var cameraView:CameraView = null;

	// The canvas on which to draw and attach movieclips to. 
	// Created and attached dynamically in the constructor.
	// The scene variable basically shows what the camera object 'sees'.
	private var scene:MovieClip;
	
	// A list of all the objects that need to be translated from 3D to 2D
	private var objectsInScene:Array;
	
	// A function that is called when the 3DScene is updated
	public var onUpdate:Function;

	// Focal length to determine perspective scaling. The smaller the 
	// number, the deeper the position, but the faster transition to the camera
	private static var FOCALLENGTH:Number = 250;

	/**
	 * Constructor: - Takes a parent (owner) and attaches an empty canvas to it.
	 *				- Initializes the array that will contain all the objects in the scene
	 *				- Initializes the camera object, and makes the 3Dscene update automatically
	 *				whenever the camera has changed.
	 */
	private function AScene3D(parent:MovieClip) {
		scene = parent.createEmptyMovieClip("scene_canvas3D", 0);
		objectsInScene = new Array();
		
		// initialize camera, add objects and update camera to display them properly
		cameraView = new CameraView(0, -300, 0);
		cameraView.addObserver(this);
		createObjects();
		cameraView.update();
	}
	
	/**
	 * Return null for this abstract singleton class.
	 */
	public static function get3DScene(parent:MovieClip): AScene3D {
		return null;
	}
	
	/**
	 * Returns the canvas of the 3Dscene.
	 */
	public function getScene():MovieClip {
		return this.scene;
	}
	
	/**
	 * Returns the camera of the 3Dscene.
	 */
	public function getCamera():CameraView {
		return AScene3D.cameraView;
	}
	
	/**
	 * All subclasses should override this method to populate the 3DScene.
	 * Alternatively, use add3DObject / remove3DObject instead to populate.
	 */
	private function createObjects() {
		//
	}
	
	/**
	 * Adds a 3D object to the 3Dscene, and directly translates it to 2D space
	 */
	public function add3DObject(object3D:Object):Void {
		if (object3D != null) {
			objectsInScene.push(object3D);
			drawObject(object3D, getCamera());
		} else {
			trace("error: invalid object");
		}
	}
	
	/**
	 * Removes a 3D object from the 3Dscene
	 */
	public function remove3DObject(Object3D:Object):Void {
		if (Object3D != null) {
			for (var i = 0; i < objectsInScene.length; i++) {
				if (objectsInScene[i] == Object3D) {
					objectsInScene.splice(i, 1);
					return;
				}
			}
		} else {
			trace("error: object to delete is [null]");
		}
		trace("error: object to delete not found");
	}

	/**
	 * Converts the x,y,z properties of an object from 3D to 2D space, 
	 * using _xscale and y_scale to emphasize depth.
	 * Also hides whatever object can't be seen by the camera.
	 * FOCALLENGTH determines how strong the 3D effect is.
	 */
	public function drawObject(object3D:Object, cameraView:CameraView) {
		var x:Number = object3D.x - cameraView.x;
		var y:Number = object3D.y - cameraView.y;
		var z:Number = object3D.z - cameraView.z;
		
		var scaleRatio = FOCALLENGTH / (FOCALLENGTH + z);
		if (scaleRatio <= 0) {
			object3D.behindCamera = true;
			object3D.visible = false;
		} else {
			object3D.behindCamera = false;
			object3D._x = x * scaleRatio;
			object3D._y = y * scaleRatio;
			object3D._visible = object3D._x > -512 && object3D._x < 512;
			object3D._visible = object3D._visible && object3D._y > -scene._y && object3D._y < 768-scene._y;
			object3D._xscale = object3D._yscale = 100 * scaleRatio;
		}
	};
	
	/**
	 * Updates all objects in the scene. 
	 * Automatically called if the camera changes. 
	 * Call drawObject if you want to update a single object only.
	 */
	public function update(o:Observable, infoObj:Object):Void {
		for (var i in objectsInScene) {
			drawObject(objectsInScene[i], CameraView(o));
		}
		if (onUpdate != null) {
			onUpdate();
		}
	}
}